// -*- mode: C++; indent-tabs-mode: nil; c-basic-offset: 4 -*-

/*
 * (c) 2017 Petr Ročkai <code@fixp.eu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once
#include <divine/vm/lx-slot.hpp>

DIVINE_RELAX_WARNINGS
#include <llvm/IR/Instructions.h>
DIVINE_UNRELAX_WARNINGS

namespace divine::vm::lx
{

enum OpCodes { OpHypercall = llvm::Instruction::OtherOpsEnd + 1, OpBB, OpArg, OpDbg, OpDbgCall };

enum Hypercall /* see divm.h for prototypes & documentation */
{
    NotHypercall = 0,
    NotHypercallButIntrinsic = 1,

    HypercallChoose,
    HypercallFault,

    HypercallTestCrit,
    HypercallTestLoop,
    HypercallTestTaint,

    HypercallCtlSet,
    HypercallCtlGet,
    HypercallCtlFlag,

    HypercallPeek,
    HypercallPoke,

    /* feedback */
    HypercallTrace,
    HypercallSyscall,

    /* memory management */
    HypercallObjMake,
    HypercallObjClone,
    HypercallObjFree,
    HypercallObjShared,
    HypercallObjResize,
    HypercallObjSize
};

enum DbgSubcode
{
    DbgValue, DbgDeclare, DbgBitCast
};

struct Instruction : _VM_Instruction
{
    Slot result() const { ASSERT( valuecount ); return value( 0 ); }

    Slot operand( int i ) const
    {
        int adjusted = ( i >= 0 ) ? ( i + 1 ) : ( i + valuecount );
        return value( adjusted );
    }

    Slot value( int i ) const
    {
        int idx = ( i >= 0 ) ? i : ( i + valuecount );
        ASSERT_LT( idx, valuecount );
        auto op = raw_value( idx );
        if ( !ext( op ) )
            return Slot( op );

        int exti = 0;
        for ( int j = 0; j < idx; ++j )
            if ( ext( raw_value( j ) ) )
                ++ exti;

        auto ext = raw_ext( valuecount + exti );
        return Slot( op, ext );
    }

    int argcount() const { return valuecount - 1; }
    bool has_result() const { return valuecount > 0; }
    static bool ext( _VM_Operand op ) { return op.type == Slot::Agg || op.type == Slot::Other; }

    _VM_Operand raw_value( int i ) const
    {
        ASSERT_LEQ( 0, i );
        ASSERT_LT( i, valuecount );
        return *reinterpret_cast< const _VM_Operand * >( this + i );
    }

    _VM_Operand &raw_value( int i )
    {
        ASSERT_LEQ( 0, i );
        ASSERT_LT( i, valuecount );
        return *reinterpret_cast< _VM_Operand * >( this + i );
    }

    _VM_OperandExt raw_ext( int i ) const
    {
        return *reinterpret_cast< const _VM_OperandExt * >( this + i );
    }

    _VM_OperandExt &raw_ext( int i )
    {
        return *reinterpret_cast< _VM_OperandExt * >( this + i );
    }

    explicit Instruction( int opc = 0, int subc = 0 )
    {
        opcode = opc;
        subcode = subc;
    }

    Instruction( const Instruction & ) = delete;
    Instruction( Instruction && ) noexcept = default;

    friend std::ostream &operator<<( std::ostream &o, const lx::Instruction &insn )
    {
        for ( int i = 0; i < insn.valuecount; ++i )
            o << insn.value( i ) << " ";
        return o;
    }

private:
    friend struct Program;
};

using Function = _VM_Function;

static_assert( sizeof( Instruction ) == sizeof( _VM_Operand ) );
static_assert( sizeof( Instruction ) == sizeof( _VM_OperandExt ) );
static_assert( sizeof( Function ) == sizeof( _VM_Function ) );

template< typename Heap >
struct Code
{
    Heap &_heap;
    HeapPointer _base;

    CodePointer advance( CodePointer pc ) { return nextpc( pc + 1 ); }
    bool valid( CodePointer ) { NOT_IMPLEMENTED();
        /* return pc.instruction() < function( pc ).instructions.size(); */ }

    CodePointer nextpc( CodePointer pc )
    {
        if ( !valid( pc ) )
            return pc;
        auto &op = instruction( pc );
        if ( valid( pc + 1 ) && ( !op.is_instruction || op.opcode == OpBB ) )
            return nextpc( pc + 1 );
        return pc;
    }

    Instruction &instruction( CodePointer )
    {
        NOT_IMPLEMENTED();
        // return function( pc ).instruction( pc );
    }

    Function &function( CodePointer )
    {
        NOT_IMPLEMENTED();
        /*
        ASSERT_LT( pc.function(), functions.size() );
        ASSERT_LT( 0, pc.function() );
        return functions[ pc.function() ];
        */
    }
};

}
