// -*- mode: C++; indent-tabs-mode: nil; c-basic-offset: 4 -*-

/*
 * (c) 2017 Petr Ročkai <code@fixp.eu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once
#include <divine/vm/heap.hpp>
#include <divine/vm/lx-slot.hpp>
#include <divine/vm/xg-code.hpp>

DIVINE_RELAX_WARNINGS
#include <llvm/IR/Value.h>
#include <llvm/IR/Module.h>
#include <llvm/IR/DataLayout.h>
DIVINE_UNRELAX_WARNINGS

namespace divine::vm::xg
{

struct Slots
{
    std::map< llvm::Constant *, lx::Slot > _const;
    std::map< llvm::Value *, lx::Slot > _local;

    llvm::DataLayout _layout;
    llvm::Module *_module;
    AddressMap &_addr;

    int _const_size = 0, _local_size;

    void add( llvm::Function *f )
    {
        _local_size = 0;
        _local.clear();
        auto pc = _addr.code( f );

        for ( auto &arg : f->args() )
            add( &arg, pc );

        for ( auto &bb : *f )
            for ( auto &i : bb )
                add( &i, pc );

        for ( auto &bb : *f )
            for ( auto &i : bb )
                for ( auto &op : i.operands() )
                    add( op.get() );
    }

    lx::Slot allocate( lx::Slot::Location l, lx::Slot::Type t, int w = 0 )
    {
        lx::Slot s( l, t, w );

        if ( l == lx::Slot::Const )
        {
            s.offset = _const_size;
            _const_size = mem::align( _const_size + s.size(), 4 );
        }

        /* TODO frame alignment */
        if ( l == lx::Slot::Local )
        {
            s.offset = _local_size;
            _local_size = mem::align( _local_size + s.size(), 4 );
        }

        return s;
    }

    void add( llvm::Value *v, CodePointer pc = CodePointer() )
    {
        auto c = llvm::dyn_cast< llvm::Constant >( v );

        if ( c && _const.count( c ) && _local.count( v ) )
            UNREACHABLE( "value both constant and mutable at the same time" );

        if ( c && !pc.null() )
            UNREACHABLE( "tried to add a constant as a mutable value" );

        if ( ( c && _const.count( c ) ) || _local.count( v ) )
            return;

        auto s = allocate( pc.null() ? lx::Slot::Const : lx::Slot::Local,
                           xg::type( v->getType() ), xg::width( _layout, v ) );

        if ( c )
            _const.emplace( c, s );
        else
            _local.emplace( v, s );

        // return s;
    }

    lx::Slot find( llvm::Value *v )
    {
        auto c = llvm::dyn_cast< llvm::Constant >( v );
        if ( c )
        {
            ASSERT( _const.count( c ) );
            return _const[ c ];
        }
        else
        {
            ASSERT( _local.count( v ) );
            return _local[ v ];
        }
    }

    size_t size() { return _const.size(); }

    Slots( llvm::Module *m, AddressMap &addr )
        : _layout( m ), _module( m ), _addr( addr )
    {
        for ( auto var = _module->global_begin(); var != _module->global_end(); ++ var )
        {
            add( &*var );
            if ( var->hasInitializer() )
                add( var->getInitializer() );
        }

    }
};

}
